@@ -14,9 +14,7 @@
  * Library General Public License for more details.
  *
  * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library. If not, see <http://www.gnu.org/licenses/>.
  */
 
 /*
@@ -2363,6 +2361,8 @@ add_credits_section (GtkAboutDialog *about,
     }
 
   /* skip one at the end */
+  label = gtk_label_new ("");
+  gtk_grid_attach (grid, label, 1, *row, 1, 1);
   (*row)++;
 }
 
@@ -2390,7 +2390,7 @@ create_credits_page (GtkAboutDialog *about)
   gtk_container_set_border_width (GTK_CONTAINER (grid), 5);
   gtk_orientable_set_orientation (GTK_ORIENTABLE (grid), GTK_ORIENTATION_VERTICAL);
   gtk_grid_set_column_spacing (GTK_GRID (grid), 8);
-  gtk_grid_set_row_spacing (GTK_GRID (grid), 12);
+  gtk_grid_set_row_spacing (GTK_GRID (grid), 2);
   gtk_widget_set_halign (grid, GTK_ALIGN_CENTER);
   gtk_widget_set_valign (grid, GTK_ALIGN_START);
   gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (sw), grid);
