@@ -639,16 +639,16 @@ gtk_menu_bar_draw (GtkWidget *widget,
   border = gtk_container_get_border_width (GTK_CONTAINER (widget));
   context = gtk_widget_get_style_context (widget);
 
-  if (get_shadow_type (GTK_MENU_BAR (widget)) != GTK_SHADOW_NONE)
-    gtk_render_background (context, cr,
-                           border, border,
-                           gtk_widget_get_allocated_width (widget) - border * 2,
-                           gtk_widget_get_allocated_height (widget) - border * 2);
+  gtk_render_background (context, cr,
+                         border, border,
+                         gtk_widget_get_allocated_width (widget) - border * 2,
+                         gtk_widget_get_allocated_height (widget) - border * 2);
 
-  gtk_render_frame (context, cr,
-                    border, border,
-                    gtk_widget_get_allocated_width (widget) - border * 2,
-                    gtk_widget_get_allocated_height (widget) - border * 2);
+  if (get_shadow_type (GTK_MENU_BAR (widget)) != GTK_SHADOW_NONE)
+    gtk_render_frame (context, cr,
+                      border, border,
+                      gtk_widget_get_allocated_width (widget) - border * 2,
+                      gtk_widget_get_allocated_height (widget) - border * 2);
 
   GTK_WIDGET_CLASS (gtk_menu_bar_parent_class)->draw (widget, cr);
 
@@ -1041,3 +1041,32 @@ gtk_menu_bar_set_child_pack_direction (GtkMenuBar       *menubar,
       g_object_notify (G_OBJECT (menubar), "child-pack-direction");
     }
 }
+
+/**
+ * gtk_menu_bar_new_from_model:
+ * @model: a #GMenuModel
+ *
+ * Creates a new #GtkMenuBar and populates it with menu items
+ * and submenus according to @model.
+ *
+ * The created menu items are connected to actions found in the
+ * #GtkApplicationWindow to which the menu bar belongs - typically
+ * by means of being contained within the #GtkApplicationWindows
+ * widget hierarchy.
+ *
+ * Returns: a new #GtkMenuBar
+ *
+ * Since: 3.4
+ */
+GtkWidget *
+gtk_menu_bar_new_from_model (GMenuModel *model)
+{
+  GtkWidget *menubar;
+
+  g_return_val_if_fail (G_IS_MENU_MODEL (model), NULL);
+
+  menubar = gtk_menu_bar_new ();
+  gtk_menu_shell_bind_model (GTK_MENU_SHELL (menubar), model, NULL, FALSE);
+
+  return menubar;
+}
